/*
==============================================================================
    Copyright (c) 1991 by Abraxas Software, Inc.
==============================================================================
 
These CodeCheck rules implement a great many of the C programming guidelines
published by Thomas Plum. Each rule in this file is preceded by a paraphrase
of the guideline in the second edition of his book on the subject. Some of
Plum's rules cannot be checked by a static source code analyzer because they
require knowing the intentions of the programmer. Others cannot be checked by
CodeCheck because the needed trigger variables have not yet been implemented.

Reference:
   "C Programming Guidelines, Second Edition for Stanadard C" by Thomas
   Plum. Published in 1989 by Plum Hall Inc, 1 Spruce Avenue, Cardiff,
   NJ 08232, USA.

The numbering scheme used for the warnings generated by these rules is keyed
to the sections of this book. The first three digits give the section number,
and the last two digits specify the index of the rule within the section. Thus
warning number 20105 refers to the fifth rule from Section 2.01.
==============================================================================
*/


#include <check.cch>


//////////////////////////////////////////////////////
///   Section 1.01   LEXICAL RULES FOR VARIABLES   ///
//////////////////////////////////////////////////////


/****************************************************************************\
    Names of variables and functions should be written all in lower case.
\****************************************************************************/

if ( dcl_any_upper )
  if ( dcl_variable || dcl_function )
    warn( 10101, "Use all lowercase letters for %s.", dcl_name() );


/****************************************************************************\
        Use 31-character limits for all names, external and internal.
\****************************************************************************/

if ( dcl_ident_length > 31 )
  warn( 10102, "Identifier %s exceeds the 31-character limit.", dcl_name() );


/****************************************************************************\
               All names should be explicitly declared.
\****************************************************************************/

if ( dcl_not_declared )
  warn( 10103, "Parameter %s needs an explicit declaration.", dcl_name() );


/****************************************************************************\
          Within each function, declare register variables first.
\****************************************************************************/

int done_with_registers;

if ( fcn_begin )
  done_with_registers = 0;

if ( dcl_local )		//  Outermost block of the function
{
  if ( dcl_storage_flags & REGISTER_SC )
  {
    if ( done_with_registers )
      warn( 10104, "Declare register variables FIRST." );
  }
  else
    done_with_registers = 1;
}


/****************************************************************************\
         External names beginning with underscore should not be used.
\****************************************************************************/

if ( dcl_underscore )
  if ( dcl_name()[0]=='_')
    if (!dcl_static)
      if ( lin_source )
        warn( 10105, "Do not use leading underscore for project identifiers." );
	

/****************************************************************************\
         Write declarations with only one variable per source line.
\****************************************************************************/

if ( lin_dcl_count )
  if ( lin_source && (lin_dcl_count > 1) )
    warn( 10106, "Use one line per declarator." );


/****************************************************************************\
         Initializers should have braces around each subaggregate.
             >>>>>  10106 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/


/****************************************************************************\
       Variable name and type should be separated by a single space.
             >>>>>  10107 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/


/****************************************************************************\
          Comments should be attached at lined-up tab positions.
\****************************************************************************/

if ( lin_has_code )
  if ( lin_has_comment )
    if ( strstr( line(), "/*" )[-1] != '\t' )
      warn( 10108, "The trailing comment should be separate from code by a tab.");

/****************************************************************************\
    Every file-level declaration must be either extern or a definition.
\****************************************************************************/

if ( dcl_global  )
  if ( ! (dcl_definition || dcl_extern || dcl_function || dcl_typedef) )
    warn( 10110, "Use \"extern\" or an initializer here." );


/****************************************************************************\
         Storage class (if any) should precede the type specifier.
\****************************************************************************/

if ( dcl_storage_first )
  warn( 10111, "Place the storage class FIRST in this declaration." );



//////////////////////////////////////////////////
///   Section 1.02   CHOOSING VARIABLE NAMES   ///
//////////////////////////////////////////////////


/****************************************************************************\
             Names should not be re-defined in inner blocks.
\****************************************************************************/

if ( dcl_hidden )
	warn( 10201, "This declaration of %s hides another.", dcl_name() );


/****************************************************************************\
       Names with length > 4 should differ by at least two characters.
             >>>>>  10202 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/


/****************************************************************************\
        Document declared names with a comment on the declaration.
\****************************************************************************/

if ( lin_dcl_count )
	if ( ! lin_has_comment )
		warn( 10203, "This declaration needs a comment." );



/////////////////////////////////////////
///   Section 1.03   STANDARD TYPES   ///
/////////////////////////////////////////


/****************************************************************************\
       The "signed" specifier may be used only in type definitions.
\****************************************************************************/

if ( dcl_signed )
  if ( !dcl_typedef)
    warn( 10301, "Only use a typedef name for signed types." );


/****************************************************************************\
       The "unsigned" specifier may be used only in type definitions.
\****************************************************************************/

if ( dcl_unsigned )
  if ( !dcl_typedef) 
    warn( 10302, "Use a typedef name for unsigned types." );


/****************************************************************************\
       The "long double" specifier may be used only in type definitions.
\****************************************************************************/

if ( dcl_base == LONG_DOUBLE_TYPE )
  if ( ! dcl_typedef )
    warn( 10303, "Use the \"ldouble\" typedef name here." );


/****************************************************************************\
       Type int may be used as for function return types, for function,
      parameters, and for register variables. No other use is appropriate.
\****************************************************************************/

if ( dcl_variable || dcl_function || dcl_parameter )
  if ( dcl_base == INT_TYPE )
    if ( ! ( dcl_unsigned || dcl_storage_flags&REGISTER_SC ) )
      if ( ! ( dcl_function || dcl_parameter) )
        warn( 10304, "Use a typedef name in this context." );


/****************************************************************************\
           Bitwise operations must be limited to unsigned data.
\****************************************************************************/

if ( op_bit_or || op_bit_and || op_bit_xor )
{
  if (strstr(op_base_name(1),"unsigned")==0||strstr(op_base_name(2),"unsigned")==0)
  {
    warn(10305, "Bitwise operation must be limited to unsigned data.");
  }
}

if ( op_bit_not )
{
  if (strstr(op_base_name(1),"unsigned")==0)
  {
    warn(10305,"Bitwise operation n=must be limited to unsigned data.");
  }
}

///////////////////////////////////////////////////////
///   Section 1.04   MAINTAINABILITY OF CONSTANTS   ///
///////////////////////////////////////////////////////


/****************************************************************************\
           Any constant other than 0 or 1 should be manifest.
\****************************************************************************/

if ( lex_not_manifest )
  if ( ! lex_initializer )
    warn( 10401, "Use a #define or enum for this constant." );


/****************************************************************************\
        Macro and enumurated constants should be all UPPERCASE.
\****************************************************************************/

if ( pp_lowercase )
  warn( 10402, "The macro %s should be spelled in UPPERCASE.", pp_name() );

if ( dcl_enum )
  if ( ! dcl_all_upper )
    warn( 10402, "The enum %s should be spelled in UPPERCASE.", dcl_name() );


/****************************************************************************\
                Define all macros before any declarations.
\****************************************************************************/

int id_found;      //  TRUE if any function or variable has
                   //  been defined in this source file.
if ( mod_begin )
  id_found = 0;

if ( dcl_variable || dcl_function )
  if ( lin_source )
    id_found = 1;

if ( pp_macro )
  if ( id_found && lin_source )
    warn( 10403, "Define macro %s before any declaration.", pp_name() );


/****************************************************************************\
        Define each macro in exactly one header or source file.
\****************************************************************************/

if ( pp_macro_dup )
  warn( 10404, "Macro %s also defined somewhere else.",pp_name() );

/****************************************************************************\
               Nonzero arguments to exit() must be manifest.
\****************************************************************************/

int i;

if ( op_call )
{
  if ( strcmp( op_function(), "exit" ) == 0 )
  {
    i = 0;
    while ( isdigit( prev_token()[i] ) )
    {
      i++;
    }
    if ( i == strlen ( prev_token() ) )
    {
      i = 0;
      while ( prev_token()[i] == '0' )
      {
        i++;
      }
      if (i != strlen ( prev_token() ) )
        warn( 10405, "Do not use nonzero non-manifest as auguments of function exit." );
    }
  }
}

/****************************************************************************\
                  Only enums may be assigned to enums.
             >>>>>  10406 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/



/////////////////////////////////////////////
///   Section 1.05   WORD AND BYTE SIZE   ///
/////////////////////////////////////////////


/****************************************************************************\
                Do not use casts which reduce precision.
\****************************************************************************/

if ( op_cast )
{
  switch ( op_base( 1 ) )
  {
    case INT_TYPE :
      if ( op_base( 2 ) < CHAR_TYPE )
        warn( 10501, "Do not reduce integer to char.");
  }
}

///////////////////////////////////////
///   Section 1.06   BYTE ORDERING  ///
///////////////////////////////////////


/****************************************************************************\
    Do not cast from (ptr to multibyte object) to (ptr to byte).
             >>>>>  10601 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/



/////////////////////////////////////////////
///   Section 1.07   CHARACTER CONSTANTS   ///
/////////////////////////////////////////////


/****************************************************************************\
    Character constants must not contain more than one character.
\****************************************************************************/

if ( lex_char_long )
  warn( 10701, "Character constant has more than one character." );



///////////////////////////////////////
///   Section 1.08   POINTER TYPES   ///
///////////////////////////////////////


/****************************************************************************\
      Do not use implicit pointer casts, except of (void *) pointers.
           >>>>>  10801 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/


/****************************************************************************\
           Pointer to function types should be typedef'ed.
\****************************************************************************/

if ( dcl_function_ptr )
  if ( !dcl_typedef )
    warn( 10802, "Use pointer to function types only in typedef."  );

/****************************************************************************\
         Do not assign a function address to a (void *) pointer.
               >>>>>  10803 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/


/****************************************************************************\
              Do not assign NULL to a pointer to function.
            >>>>>  10804 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/


//////////////////////////////////////////////
///   Section 1.09   POINTER CONVERSIONS   ///
//////////////////////////////////////////////


/****************************************************************************\
     Do not use any pointer conversion except an assignment of NULL or
     (void *) to (data *). 
            >>>>>  10901 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/


/****************************************************************************\
Any function parameter of type size_t must be called with the correct type. 
            >>>>>  10902 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/



///////////////////////////////////////////////
///   Section 1.10   ALLOCATION INTEGRITY   ///
///////////////////////////////////////////////


/****************************************************************************\
            Do not assign a local address to a global pointer. 
             >>>>>  11001 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/


/********************                ********************************************************\
          The function free() should be embedded in a macro. 
\****************************************************************************/

if ( identifier("free") )
  if ( ! lex_macro_token )
    warn( 11002, "Embed calls to free() in a macro." );



/////////////////////////////////////
///   Section 1.11   STRUCTURES   ///
/////////////////////////////////////


/****************************************************************************\
             Do not spell tag names with leading underscores. 
\****************************************************************************/

if ( tag_begin )
  if ( tag_name()[0]=='_' )
    warn( 11101, "Do not spell tag name %s with leading underscore(s).",tag_name() );

/****************************************************************************\
            Do not spell member names with leading underscores.
\****************************************************************************/

if ( dcl_member )
  if ( dcl_name()[0]=='_' )
    warn( 11102, "Do not spell member name %s with leading underscore(s).",dcl_name() );

/****************************************************************************\
              Bitfields must be no more than 16 bits wide. 
\****************************************************************************/

if ( dcl_bitfield_size > 16 )
  warn( 11103, "Bitfield width exceeds 16." );



/////////////////////////////////////
///   Section 1.12   STRUCTURES   ///
/////////////////////////////////////


/****************************************************************************\
         Do not pass a pointer-to-constant-data to a function
           whose prototype allows the data to be modified.
            >>>>>  11201 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/


/****************************************************************************\
            If an initializer is a string literal, then the
             variable must be a pointer to constant data.
\****************************************************************************/

int  string_initializer;

if ( lex_initializer == 5 )
	string_initializer = 1;

if ( dcl_initializer )
{
  if ( string_initializer && ((dcl_level_flags(0)&CONST_FLAG == 0)))
    warn( 11202, "Variable \"%s\" should be declared const.", dcl_name() );
  string_initializer = 0;
}



//////////////////////////////////////////////////////
///   Section 2.01   LEXICAL RULES FOR OPERATORS   ///
//////////////////////////////////////////////////////


/****************************************************************************\
            High precedence operators should not be separated
                  from their operands by whitespace.
\****************************************************************************/


if ( op_high_infix )
{
  if ( op_white_before )
    warn( 20101, "Remove the space before this operator." );
  if ( op_white_after )
    warn( 20102, "Remove the space after this operator." );
}

if ( op_high_prefix )
  if ( op_white_before )
    warn( 20101, "Remove the space before this operator." );

if ( op_high_postfix )
  if ( op_white_after )
    warn( 20102, "Remove the space after this operator." );


/****************************************************************************\
            Low precedence operators should be separated
                from their operands by whitespace.
\****************************************************************************/

if ( op_low || op_colon_2 )
{
  if ( ! op_white_before )
    warn( 20103, "Insert space before this operator." );
  if ( ! op_white_after )
    warn( 20104, "Insert space after this operator." );
}


/****************************************************************************\
    Expressions within parentheses should be written with no space after
    the opening parenthesis and no space before the closing parenthesis.
\****************************************************************************/

if ( op_open_paren )
  if ( op_white_after )
    warn( 20105, "Remove the space after this parenthesis." );

if ( op_close_paren )
  if ( op_white_before )
    warn( 20106, "Remove the space before this parenthesis." );


/****************************************************************************\
            Commas and semicolons should be followed by whitespace.
\****************************************************************************/

if ( lex_punct_after )
  warn( 20107, "Use whitespace after commas and semicolons." );


/****************************************************************************\
       Medium-precedence operators with no space around them must bind
          their operands more tightly than the adjacent operators.
           >>>>>  20108 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/


/****************************************************************************\
              Keywords should be followed by one space.
\****************************************************************************/

if ( lex_key_no_space )
  warn( 20109, "This keyword must be followed by a space." );



///////////////////////////////////////////
///   Section 2.02   EVALUATION ORDER   ///
///////////////////////////////////////////


/****************************************************************************\
          Avoid dependence upon the evaluation order of expressions
                    between the sequence points of C.          
               >>>>>  20201 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/



//////////////////////////////////////
///   Section 2.03   PARENTHESES   ///
//////////////////////////////////////


/****************************************************************************\
               >>>>>  20301 IS NOT YET IMPLEMENTED  <<<<<
\****************************************************************************/
